@@ -7,12 +7,12 @@
 //! between API calls to reduce heap allocations. In some cases the return value of an API may
 //! borrow from the scratch buffer.
 //!
-//! = Important =
+//! # Important
 //! You *must* depend on [`oc-wasm-futures`](https://gitlab.com/Hawk777/oc-wasm-futures) with the
 //! `proper-waker` feature in your own application if your chosen executor requires the
 //! `proper-waker` feature.
 //!
-//! = Example =
+//! # Example
 //! ```
 //! extern crate alloc;
 //! use alloc::boxed::Box;
@@ -90,7 +90,7 @@
 //! }
 //!
 //! #[no_mangle]
-//! fn run(_: i32) -> i32 {
+//! pub extern "C" fn run(_: i32) -> i32 {
 //! 	static mut PANIC_HOOK_SET: bool = false;
 //! 	static mut EXECUTOR: OnceCell<Cassette<Pin<Box<dyn Future<Output = ()>>>>> = OnceCell::new();
 //!
@@ -160,6 +160,7 @@ pub mod error;
 pub mod filesystem;
 pub mod gpu;
 pub mod inventory;
+pub mod keyboard;
 pub mod redstone;
 pub mod robot;
 pub mod screen;
